/*
 * Controller.java
 */
 
package de.ba_ravensburg.gdv.controller;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;

import javax.swing.JButton;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import de.ba_ravensburg.gdv.model.Scene;
import de.ba_ravensburg.gdv.view.ControlPanel;
import de.ba_ravensburg.gdv.view.MainUI;

/**
 * The <i>Controller</i> creates view and model and contains all functionality
 * that can be called through the user interface.
 * 
 * @author Max Schloss
 * @author Ansgar Wiechers
 */
public class Controller
  implements ActionListener, WindowListener, ChangeListener {

  private final double v = 2.5; // velocity in m/s
  private final double r = 0.5; // turn radius of the camera
  private final double fps = 25.0;
  private long sleepTime = 1;
  private final double STEP = ControlPanel.STEP;
  private final int X = Scene.X;
  private final int Y = Scene.Y;
  private final int Z = Scene.Z;
  
  private MainUI view;
  private Scene model;
  private RenderPipeline pipeline;
  
  /**
   * Constructs a new controller which then creates the model and the view.
   */
  public Controller() {
    model = new Scene(0.0, 0.0, 10.0);
    view  = new MainUI(this);
    pipeline = new RenderPipeline();
    renderScene();
    refreshControlPanel();
  }
  
  /**
   * Program entry point. The sole purpose of this function is to create a new
   * controller when the program is started.
   * 
   * @param args arguments the program was started with.
   */
  public static void main(String[] args) {
    Controller controller = new Controller();
  }
  
  /**
   * 
   */
  private void refreshControlPanel() {
    double[] camPos = model.getPos();
    view.getViewControl().getSmX().setValue(new Double(camPos[0]));
    view.getViewControl().getSmY().setValue(new Double(camPos[1]));
    view.getViewControl().getSmZ().setValue(new Double(camPos[2]));
    view.getViewControl().getSmZoom().setValue(new Double(model.getAperture() *
                                                          2.0));
  }

  public void actionPerformed(ActionEvent e) {
    int index;
    
    Object source = e.getSource();
    if (source instanceof JButton) {
      JButton pressedButton = (JButton) source;
      
      if (pressedButton == view.getViewControl().getBtStartSeq()) {
        index = view.getViewControl().getCbSequenzAuswahl().getSelectedIndex();
        if (index == 0)
          cameraSequenceCircular();
        else
          cameraSequenceLinear();
        
        refreshControlPanel();
      } else {
        // Check which Button was pressed
        if (pressedButton == view.getViewControl().getBtVor()) {
            model.moveForward(STEP);
        } else if (pressedButton == view.getViewControl().getBtRechts()) {
            model.moveRight(STEP);
        } else if (pressedButton == view.getViewControl().getBtZurueck()) {
            model.moveBackward(STEP);
        } else if (pressedButton == view.getViewControl().getBtLinks()) {
            model.moveLeft(STEP);
        } else if (pressedButton == view.getViewControl().getBtHoch()) {
            model.moveUpward(STEP);
        } else if (pressedButton == view.getViewControl().getBtRunter()) {
            model.moveDownward(STEP);
        } else if (pressedButton == view.getViewControl().getBtDrehLinks()) {
            model.turnLeft(STEP);
        } else if (pressedButton == view.getViewControl().getBtDrehRechts()) {
            model.turnRight(STEP);
        } else if (pressedButton == view.getViewControl().getBtNickHoch()) {
            model.turnUpward(STEP);
        } else if (pressedButton == view.getViewControl().getBtNickRunter()) {
            model.turnDownward(STEP);
        } else if (pressedButton == view.getViewControl().getBtGier()) {
            try {
              double gier = Double.parseDouble(
                              view.getViewControl().getTfGier().getText());
              double nick = Double.parseDouble(
                              view.getViewControl().getTfNick().getText());
              model.setDirection(gier, nick);
            } catch (NumberFormatException ex) {
            } finally {
              view.getViewControl().getTfNick().setText("0"); //$NON-NLS-1$
              view.getViewControl().getTfGier().setText("0"); //$NON-NLS-1$
            }
            
            
        }
        renderScene();
        refreshControlPanel();
      }
    }
  }

  public void stateChanged(ChangeEvent e) {
    Object source = e.getSource();
    if (source instanceof JSpinner) {
      SpinnerNumberModel spModel =
        (SpinnerNumberModel)((JSpinner)source).getModel();
      
      double[] camPos = model.getPos();

      // Check witch SpinnerModel was changed
      if (spModel == view.getViewControl().getSmX()) {
        model.setPos(((Double) spModel.getValue()).doubleValue(), camPos[1],
                     camPos[2]);
      } else if (spModel == view.getViewControl().getSmY()) {
        model.setPos(camPos[0], ((Double) spModel.getValue()).doubleValue(),
                     camPos[2]);
      } else if (spModel == view.getViewControl().getSmZ()) {
        model.setPos(camPos[0], camPos[1],
                     ((Double) spModel.getValue()).doubleValue());
      } else if (spModel == view.getViewControl().getSmGier()) {
      } else if (spModel == view.getViewControl().getSmNick()) {
      } else if (spModel == view.getViewControl().getSmZoom()) {
        model.setAperture(((Double) spModel.getValue()).doubleValue() / 2.0);
      }
      renderScene();
    }
  }
  
  // Unused WindowListener-Events
  public void windowClosing(WindowEvent e) {}
  public void windowDeactivated(WindowEvent e) {}
  public void windowActivated(WindowEvent e) {}
  public void windowClosed(WindowEvent e) {}
  public void windowDeiconified(WindowEvent e) {}
  public void windowIconified(WindowEvent e) {}
  public void windowOpened(WindowEvent e) {}
  
  /**
   * Initiate the rendering of the scene.
   */
  private void renderScene() {
    pipeline.put(model.getTriangles(), model.getAperture());
    view.getView3D().paint(pipeline.get());
    view.getOverview().paint(model.getPos(),
                             model.getAperture(),
                             model.getDirection());
  }
  
  /**
   * Move the camera along a circular trail with radius r&nbsp;=&nbsp;90&nbsp;m
   * and centre (50;100;10). The camera is always facing towards the centre of
   * the circle, the z-coordinate is always z0&nbsp;=&nbsp;10&nbsp;m. The motion
   * stops after the camera has finished one complete circle.
   */
  private void cameraSequenceCircular() {
    double r = 90.0;
    double x0 = 50.0, y0 = 100.0, z0 = 10.0;
    double phi, currentAngle;
    
    model.setPos(x0 + r, y0, z0);
    model.setDirection(-1.0, 0.0, 0.0);
    
    /* U = 2 * Pi * r
     * b = U / v * fps
     * phi = 360 / b
     * 
     * => phi = 360 / (2 * Pi * r * fps / v)
     */
    phi = 180.0 * v / Math.PI / r / fps;
    
    for (currentAngle = 0.0; currentAngle < 360.0; currentAngle += phi) {
      model.setPos(x0 + r * Math.cos(Math.toRadians(currentAngle)),
                   y0 + r * Math.sin(Math.toRadians(currentAngle)),
                   z0);
      
      model.turnLeft(phi);
      renderScene();
      sleep(sleepTime);
    }
  }
  
  /**
   * Move the camera along a trail defined by the points (80;0;5), (70;20;5),
   * (97;80;5), (80;120;5), (20;180;5). Starting point is (80;0;5). Generally
   * the camera is facing towards the next point on the trail, except when
   * changing the motion direction. When the direction has to be changed
   * (usually on the fixed points of the trail), the trail is approximated by
   * an arc. Therefore the camera will in fact bypass all fixed points of the
   * trail, except for the starting and end point.
   */
  private void cameraSequenceLinear() {
    final double zPos = 5.0;
    double[] currentPos = {80.0, 0.0, zPos};
    double[][] pointList = {new double[] {70.0, 20.0, zPos},
                            new double[] {97.0, 80.0, zPos},
                            new double[] {80.0, 120.0, zPos},
                            new double[] {20.0, 180.0, zPos}};
    double[] arcStartingPoint = new double[2];
    double[] arcEndPoint = new double[2];
    double[] arcCentre = new double[2];
    double[] a = new double[2];
    double[] b = new double[2];
    double step, deltaX, deltaY, abs, k, phi, trail, dir,
           arcStep, startAngle, endAngle, currentAngle;
    int i;
    
    model.setPos(currentPos[X], currentPos[Y], currentPos[Z]);
    model.setDirection(pointList[0][X] - currentPos[X],
                       pointList[0][Y] - currentPos[Y],
                       pointList[0][Z] - currentPos[Z]);
    step = v / fps;
    
    renderScene();
    
    /* Because all points of this sequence are in the same plane, the
     * calculation of the trail can be simplified to 2D.
     * 
     * The starting and end points of the transition arc between two sections
     * of the trail are calculated as follows:
     * 
     * p:  location vector of the middle point of the trail sections.
     * a:  unit vector of the first section (aX; aY).
     * a': vector at right angles with a (aY; -aX).
     * b:  unit vector of the second section (bX; bY).
     * b': vector at right angles with b (bY; -bX).
     * r:  turn radius.
     * k:  distance of the starting and end point from p.
     * 
     *    p + k * b + r * b' = p - k * a + r * a'
     * =>     k * b + r * b' = -k * a + r * a'
     * =>        k * (a + b) = r * (a' - b')
     * =>      k * (aX + bX) = r * (aY - bY)
     *             (aY + bY)       (bX - aX)
     * =>      k * (aX + bX) = r * (aY - bY)
     *         k * (aZ + bY) = r * (bX - aX)
     * 
     * To determine k, only one of these two equations is needed.
     * 
     *         aY - bY
     * k = r * -------
     *         aX + bX
     * 
     * If k is greater than zero, then the arc is rightbound. If it is less
     * than zero, the arc is leftbound.
     */
    for (i = 0; i < pointList.length - 1; ++i) {
      dir = 1.0;
      
      deltaX = pointList[i][X] - currentPos[X];
      deltaY = pointList[i][Y] - currentPos[Y];
      abs = Math.sqrt(deltaX * deltaX + deltaY * deltaY);
      a[X] = deltaX / abs;
      a[Y] = deltaY / abs;
      
      deltaX = pointList[i+1][X] - pointList[i][X];
      deltaY = pointList[i+1][Y] - pointList[i][Y];
      abs = Math.sqrt(deltaX * deltaX + deltaY * deltaY);
      b[X] = deltaX / abs;
      b[Y] = deltaY / abs;
      
      k = r * (a[Y] - b[Y]) / (a[X] + b[X]);
      
      phi = Math.toDegrees(Math.acos((a[X] * b[X] + a[Y] * b[Y]) /
                                     Math.sqrt((a[X] * a[X] + a[Y] * a[Y]) *
                                               (b[X] * b[X] + b[Y] * b[Y]))));
      
      if (k < 0.0) {
        // The arc is leftbound, so we have to correct phi and k.
        k = -k;
        phi = -phi;
        dir = -1.0;
      }
      
      arcStartingPoint[X] = pointList[i][X] - k * a[X];
      arcStartingPoint[Y] = pointList[i][Y] - k * a[Y];
      arcEndPoint[X] = pointList[i][X] + k * b[X];
      arcEndPoint[Y] = pointList[i][Y] + k * b[Y];
      arcCentre[X] = pointList[i][X] - k * a[X] + dir * r * a[Y];
      arcCentre[Y] = pointList[i][Y] - k * a[Y] - dir * r * a[X];
      
      // Move to starting point of arc
      deltaX = arcStartingPoint[X] - currentPos[X];
      deltaY = arcStartingPoint[Y] - currentPos[Y];
      abs = Math.sqrt(deltaX * deltaX + deltaY * deltaY);
      
      for (trail = 0.0; trail < abs; trail += step) {
        model.moveForward(step);
        renderScene();
        sleep(sleepTime);
      }
      
      model.setPos(pointList[i][X], pointList[i][Y], zPos);
      renderScene();
      
      // Move through arc.
      arcStep = 45.0 * step / r / Math.PI;
        
      if (phi < 0.0) {
        // Arc is leftbound.
        startAngle = Math.toDegrees(Math.atan2(-a[X], a[Y]));
        endAngle = startAngle - phi;
        
        for (currentAngle = startAngle; currentAngle < endAngle;
             currentAngle += arcStep) {
          model.setPos(arcCentre[X] + r *
                         Math.cos(Math.toRadians(currentAngle)),
                       arcCentre[Y] + r *
                         Math.sin(Math.toRadians(currentAngle)),
                       zPos);
          model.turnLeft(arcStep);
          renderScene();
          sleep(sleepTime * 30);
        }
      } else {
        // Arc is rightbound.
        startAngle = Math.toDegrees(Math.atan2(a[X], -a[Y]));
        if (startAngle < 0.0)
          startAngle += 360.0;
        endAngle = startAngle - phi;
        
        for (currentAngle = startAngle; currentAngle > endAngle;
             currentAngle -= arcStep) {
          model.setPos(arcCentre[X] + r * Math.cos(Math.toRadians(currentAngle)),
                       arcCentre[Y] + r * Math.sin(Math.toRadians(currentAngle)),
                       zPos);
          model.turnRight(arcStep);
          renderScene();
          sleep(sleepTime * 30);
        }
      }
      
      model.setPos(arcEndPoint[X], arcEndPoint[Y], zPos);
      model.setDirection(b[X], b[Y], 0.0);
      renderScene();
      
      currentPos[X] = arcEndPoint[X];
      currentPos[Y] = arcEndPoint[Y];
    }
    
    // Move towards end point.
    deltaX = pointList[pointList.length-1][X] - currentPos[X];
    deltaY = pointList[pointList.length-1][Y] - currentPos[Y];
    abs = Math.sqrt(deltaX * deltaX + deltaY * deltaY);
      
    for (trail = 0.0; trail < abs; trail += step) {
      model.moveForward(step);
      renderScene();
      sleep(sleepTime);
    }
      
    model.setPos(pointList[i][X], pointList[i][Y], zPos);
    renderScene();
  }
  
  /**
   * Sleep for given amount of time.
   * 
   * @param millis time to sleep in milliseconds
   */
  private void sleep(long millis) {
    try {
      Thread.sleep(millis);
    } catch (InterruptedException e) {
      // sleep() interrupted. Back to work.
    }
  }

}
