/*
 * Scene.java
 */
 
package de.ba_ravensburg.gdv.model;

import java.awt.Color;
import java.util.HashSet;
import java.util.Iterator;

/**
 * The <i>Scene</i> contains all triangles as well as the camera. The camera is
 * defined by its position, direction and aperture. The height of the camera
 * position (z-coordinate) cannot be less than zero, the aperture must be
 * greater than zero, but at most 180.
 * 
 * @author Ansgar Wiechers
 */
public class Scene {
  
  /**
   * The dimension of the Scene.
   */
  public static final int DIMENSION = 3; // Dimension of point coordinates.
  
  /**
   * Constant representing x-direction.
   */
  public static final int X = 0;
  
  /**
   * Constant representing y-direction.
   */
  public static final int Y = 1;
  
  /**
   * Constant representing z-direction.
   */
  public static final int Z = 2;
  
  /**
   * The default color of the scene's base.
   */
  public static final Color GROUND_COLOR = new Color(102, 153, 102);
  
  /**
   * Constant representing vertical turns.
   */
  static final int UPDOWN = 3;
  
  /**
   * Constant representing horizontal turns.
   */
  static final int LEFTRIGHT = 4;
  
  /**
   * The default position of the camera.
   */
  static final double[] DEFAULT_POSITION = {0.0, 0.0, 0.0};
  
  /**
   * The default viewing direction of the camera.
   */
  static final double[] DEFAULT_DIRECTION = {0.0, 1.0, 0.0};
  
  /**
   * The default aperture of the camera.
   */
  static final double DEFAULT_APERTURE = 30.0; // degree
  
  /**
   * The default direction of the x-axis of the camera coordinate.
   */
  static final double[] DEFAULT_CAMERA_X = {1.0, 0.0, 0.0};
  
  /**
   * The default direction of the y-axis of the camera coordinate. This is the
   * viewing direction of the camera.
   */
  static final double[] DEFAULT_CAMERA_Y = {DEFAULT_DIRECTION[X],
                                            DEFAULT_DIRECTION[Y],
                                            DEFAULT_DIRECTION[Z]};
  
  /**
   * The default direction of the z-axis of the camera coordinate.
   */
  static final double[] DEFAULT_CAMERA_Z = {0.0, 0.0, 1.0};
  
  private HashSet<Triangle> triangles;
  private double[] cameraPos = new double[DIMENSION];
  private double[] cameraX = {1.0, 0.0, 0.0};
  private double[] cameraY = {0.0, 1.0, 0.0};
  private double[] cameraZ = {0.0, 0.0, 1.0};
  private double aperture;
  
  /**
   * Constructs a new scene with default camera position, camera direction and
   * camera aperture.
   * 
   * @throws IllegalArgumentException
   */
  public Scene() throws IllegalArgumentException {
    this(DEFAULT_POSITION[X], DEFAULT_POSITION[Y], DEFAULT_POSITION[Z]);
  }
  
  /**
   * Constructs a new scene with given camera position and default camera
   * direction and camera aperture.
   * 
   * @param x x-coordinate of the camera position in the world coordinate.
   * @param y y-coordinate of the camera position in the world coordinate.
   * @param z z-coordinate of the camera position in the world coordinate.
   * @throws IllegalArgumentException
   */
  public Scene(double x, double y, double z) throws IllegalArgumentException {
    this(x, y, z, DEFAULT_DIRECTION[X], DEFAULT_DIRECTION[Y],
         DEFAULT_DIRECTION[Z]);
  }
  
  /**
   * Constructs a new scene with given camera position and camera direction and
   * default camera aperture.
   * 
   * @param x x-coordinate of the camera position in the world coordinate.
   * @param y y-coordinate of the camera position in the world coordinate.
   * @param z z-coordinate of the camera position in the world coordinate.
   * @param dirX x-coordinate of initial view direction of the camera.
   * @param dirY y-coordinate of initial view direction of the camera.
   * @param dirZ z-coordinate of initial view direction of the camera.
   * @throws IllegalArgumentException
   */
  public Scene(double x, double y, double z, double dirX, double dirY,
               double dirZ) throws IllegalArgumentException {
    triangles = new HashSet<Triangle>();
    
    setPos(x, y, z);
    setDirection(dirX, dirY, dirZ);
    aperture = DEFAULT_APERTURE;
    
    loadScene();
  }
  
  /**
   * Constructs a new scene with given camera position and camera direction and
   * default camera aperture.
   * 
   * @param x x-coordinate of the camera position in the world coordinate.
   * @param y y-coordinate of the camera position in the world coordinate.
   * @param z z-coordinate of the camera position in the world coordinate.
   * @param phi x-coordinate of initial view direction of the camera.
   * @param rho y-coordinate of initial view direction of the camera.
   * @throws IllegalArgumentException
   */
  public Scene(double x, double y, double z, double phi, double rho)
    throws IllegalArgumentException {
      triangles = new HashSet<Triangle>();
    
      setPos(x, y, z);
      setDirection(phi, rho);
      aperture = DEFAULT_APERTURE;
    
      loadScene();
  }
  
  /**
   * Loads triangles into the scene.
   */
  private void loadScene() {
    // base
    add(new Triangle(-50.0, -50.0, 0.0, -50.0, 250.0, 0.0, 250.0, 250.0, 0.0,
                     GROUND_COLOR));
    add(new Triangle(-50.0, -50.0, 0.0, 250.0, 250.0, 0.0, 250.0, -50.0, 0.0,
                     GROUND_COLOR));
    
    // cube
    add(new Triangle(40.0, 40.0, 0.0, 40.0, 60.0, 0.0, 40.0, 60.0, 20.0,
                     new Color(200, 0, 0)));
    add(new Triangle(40.0, 40.0, 0.0, 40.0, 60.0, 20.0, 40.0, 40.0, 20.0,
                     new Color(200, 0, 0)));
    add(new Triangle(40.0, 40.0, 0.0, 40.0, 40.0, 20.0, 60.0, 40.0, 20.0,
                     new Color(180, 0, 0)));
    add(new Triangle(40.0, 40.0, 0.0, 60.0, 40.0, 20.0, 60.0, 40.0, 0.0,
                     new Color(180, 0, 0)));
    add(new Triangle(60.0, 40.0, 0.0, 60.0, 40.0, 20.0, 60.0, 60.0, 20.0,
                     new Color(160, 0, 0)));
    add(new Triangle(60.0, 40.0, 0.0, 60.0, 60.0, 20.0, 60.0, 60.0, 0.0,
                     new Color(160, 0, 0)));
    add(new Triangle(60.0, 60.0, 0.0, 60.0, 60.0, 20.0, 40.0, 60.0, 20.0,
                     new Color(140, 0, 0)));
    add(new Triangle(60.0, 60.0, 0.0, 40.0, 60.0, 20.0, 40.0, 60.0, 0.0,
                     new Color(140, 0, 0)));
    add(new Triangle(40.0, 40.0, 20.0, 40.0, 60.0, 20.0, 60.0, 60.0, 20.0,
                     new Color(120, 0, 0)));
    add(new Triangle(40.0, 40.0, 20.0, 60.0, 60.0, 20.0, 60.0, 40.0, 20.0,
                     new Color(120, 0, 0)));
    
    // pyramid
    add(new Triangle(50.0, 175.0, 0.0, 50.0, 195.0, 0.0, 65.0, 185.0, 50.0,
                     new Color(0, 0, 200)));
    add(new Triangle(50.0, 175.0, 0.0, 65.0, 185.0, 50.0, 80.0, 175.0, 0.0,
                     new Color(0, 0, 180)));
    add(new Triangle(80.0, 175.0, 0.0, 65.0, 185.0, 50.0, 80.0, 195.0, 0.0,
                     new Color(0, 0, 160)));
    add(new Triangle(80.0, 195.0, 0.0, 65.0, 185.0, 50.0, 50.0, 195.0, 0.0,
                     new Color(0, 0, 140)));
    
    // triangular column
    add(new Triangle(110.0, 110.0, 0.0, 90.0, 130.0, 0.0, 90.0, 130.0, 40.0,
                     new Color(200, 200, 0)));
    add(new Triangle(110.0, 110.0, 0.0, 90.0, 130.0, 40.0, 110.0, 110.0, 40.0,
                     new Color(200, 200, 0)));
    add(new Triangle(110.0, 110.0, 0.0, 110.0, 110.0, 40.0, 120.0, 140.0, 40.0,
                     new Color(180, 180, 0)));
    add(new Triangle(110.0, 110.0, 0.0, 120.0, 140.0, 40.0, 120.0, 140.0, 0.0,
                     new Color(180, 180, 0)));
    add(new Triangle(120.0, 140.0, 0.0, 120.0, 140.0, 40.0, 90.0, 130.0, 40.0,
                     new Color(160, 160, 0)));
    add(new Triangle(120.0, 140.0, 0.0, 90.0, 130.0, 40.0, 90.0, 130.0, 0.0,
                     new Color(160, 160, 0)));
    add(new Triangle(110.0, 110.0, 40.0, 90.0, 130.0, 40.0, 120.0, 140.0, 40.0,
                     new Color(140, 140, 0)));
  }
  
  /**
   * Add a triangle to the scene.
   * 
   * @param theTriangle triangle to be added to the scene.
   * @return true if the triangle was successfully added, else false.
   */
  public boolean add(Triangle theTriangle) {
    double[] point;
    
    point = theTriangle.getPointA();
    point[X] -= cameraPos[X];
    point[Y] -= cameraPos[Y];
    point[Z] -= cameraPos[Z];
    
    point = theTriangle.getPointB();
    point[X] -= cameraPos[X];
    point[Y] -= cameraPos[Y];
    point[Z] -= cameraPos[Z];
    
    point = theTriangle.getPointC();
    point[X] -= cameraPos[X];
    point[Y] -= cameraPos[Y];
    point[Z] -= cameraPos[Z];
    
    return triangles.add(theTriangle);
  }
  
  /**
   * Get all triangles from the scene.
   * 
   * @return collection with all triangles of the scene.
   */
  public HashSet getTriangles() {
    return triangles;
  }
  
  /**
   * Get the current aperture of the camera.
   * 
   * @return current aperture of the camera in degree.
   */
  public double getAperture() {
    return aperture;
  }
  
  /**
   * Get the current camera position in the world coordinate.
   * 
   * @return current position of the camera.
   */
  public double[] getPos() {
    return cameraPos;
  }
  
  /**
   * Get the view direction of the camera. This returns just the viewing
   * direction (the y-axis of the camera coordinate), not the entire orientation
   * of the camera.
   * 
   * @return view direction of the camera.
   */
  public double[] getDirection() {
    return cameraY;
  }
  
  /**
   * Set the aperture of the camera.
   * <p>
   * <table><tr>
   * <td>Pre:  <td>0 &lt; theAperture &lt;= 180
   * </tr><tr>
   * <td>Post: <td>aperture == theAperture
   * </tr></table>
   * 
   * @param theAperture the new aperture.
   * @throws IllegalArgumentException
   */
  public void setAperture(double theAperture) throws IllegalArgumentException {
    if (theAperture <= 0.0 || theAperture > 180.0)
      throw new IllegalArgumentException();
    
    aperture = theAperture;
  }

  /**
   * Set camera position in the scene.
   * <p>
   * <table><tr>
   * <td>Pre:  <td>z &gt;= 0
   * </tr><tr>
   * <td>Post: <td>The position of the camera is set to cameraPosition
   * </tr></table>
   * @param x x-coordinate of the new camera position.
   * @param y y-coordinate of the new camera position.
   * @param z z-coordinate of the new camera position.
   * @throws IllegalArgumentException
   */
  public void setPos(double x, double y, double z)
    throws IllegalArgumentException {
    Triangle triangle;
    double dx, dy, dz;
    double[] direction = new double[DIMENSION];
    
    if (z < 0.0)
      throw new IllegalArgumentException();
    
    dx = x - cameraPos[X];
    dy = y - cameraPos[Y];
    dz = z - cameraPos[Z];
    
    /* Transform the translation vector (dx;dy;dz) into the camera coordinate.
     * The transformation matrix T is the transposed matrix C, which consists
     * of the base vectors of the camera coordinate.
     * 
     * E = world coordinate (identity matrix)
     * C = camera coordinate
     * T = transformation matrix
     * t = translation vector in E (dx;dy;dz)
     * r = translation vector in C (direction)
     * 
     * T = C' * E = C'
     * 
     *          ( cameraX[X] cameraX[Y] cameraX[Z] )
     * T = C' = ( cameraY[X] cameraY[Y] cameraY[Z] )
     *          ( cameraZ[X] cameraZ[Y] cameraZ[Z] )
     *
     * r = T * t = C' * t
     */ 
    direction[X] = cameraX[X] * dx + cameraX[Y] * dy + cameraX[Z] * dz;
    direction[Y] = cameraY[X] * dx + cameraY[Y] * dy + cameraY[Z] * dz;
    direction[Z] = cameraZ[X] * dx + cameraZ[Y] * dy + cameraZ[Z] * dz;
    
    Iterator it = triangles.iterator();
    while (it.hasNext()) {
      triangle = (Triangle)it.next();
      triangle.move(-direction[X], X);
      triangle.move(-direction[Y], Y);
      triangle.move(-direction[Z], Z);
    }
    
    cameraPos[X] = x;
    cameraPos[Y] = y;
    cameraPos[Z] = z;
  }

  /**
   * Set the new direction of the camera.
   * <p>
   * <table><tr>
   * <td>Pre:  <td>at least one of <i>x</i>, <i>y</i> and <i>z</i> is different
   *               from zero.
   * </tr><tr>
   * <td>Post: <td>The view direction of the camera is changed to the new
   *               coordinates. The scene is rotated accordingly.
   * </tr></table>
   * 
   * @param x x-coordinate of the new camera direction.
   * @param y y-coordinate of the new camera direction.
   * @param z z-coordinate of the new camera direction.
   * @throws IllegalArgumentException
   */
  public void setDirection(double x, double y, double z)
    throws IllegalArgumentException {
    double phi = 0.0, rho = 90.0, sqrSumXY;
    
    if (x == 0.0 && y == 0.0 && z == 0.0)
      throw new IllegalArgumentException();
    
    /* To change the view to the new direction, the scene needs to be turned
     * by the angles phi and rho.
     * 
     * r:   (x;y;z)
     * phi: smaller angle between the projection of r into the xy-plane and
     *      the positive y-axis.
     * rho: angle between r and its projection into the yz-plane.
     * 
     * If both r[X] and r[Y] are zero, then the new direction is at right
     * angles with the xy-plane.
     * => - phi is 0 (the default value).
     *    - rho is 90 (the default value) or -90, depending on the
     *      algebraic signedness of the z-coordinate.
     * 
     * Otherwise the new direction is not at right angles with the xy-plane.
     * => - phi is out of ]-180;180].
     *    - rho is out of [-90;90].
     */
    if (x != 0.0 || y != 0.0) {
      // If x is zero and y is greater than zero, then phi is 0 (the default
      // value).
      if (x != 0.0 || y <= 0.0) {
        phi = Math.atan2(y, x) / Math.PI * 180.0;
        if (x > 0.0 || y >= 0.0)
          phi -= 90.0;
        else if (y < 0.0)
          phi += 270.0;
      }

      if (z == 0.0) {
        rho = 0.0;
      } else {
        sqrSumXY = x * x + y * y;
        rho = acos(sqrSumXY / Math.sqrt((sqrSumXY + z * z) * sqrSumXY));
      }
    }
    
    // The signedness of rho is determinded by the signedness of the
    // z-coordinate.
    rho *= sgn(z);
    
    setDirection(phi, rho);
  }
  
  /**
   * Set the new direction of the camera.
   * <p>
   * <table><tr>
   * <td>Pre:  <td>true
   * </tr><tr>
   * <td>Post: <td>The view direction of the camera is changed according to the
   *               given angles. The scene is rotated accordingly.
   * </tr></table>
   * 
   * @param yawAngle   horizontal spheric coordinate angle of the new camera
   *                   direction.
   * @param pitchAngle vertical spheric coordinate angle of the new camera
   *                   direction.
   */
  public void setDirection(double yawAngle, double pitchAngle) {
    resetCamera();
    
    if (yawAngle == 0.0 && pitchAngle == 0.0)
      return;
    
    turnLeft(yawAngle);
    turnUpward(pitchAngle);
  }
  
  /**
   * Turn the camera to the left.
   * 
   * @param angle the angle in degree.
   */
  public void turnLeft(double angle) {
    if (angle != 0.0) {
      Iterator it = triangles.iterator();
      while (it.hasNext()) {
        ((Triangle)it.next()).turn(angle, LEFTRIGHT);
      }
      
      turnCameraBase(-angle, LEFTRIGHT);
    }
  }
  
  /**
   * Turn the camera to the right.
   * 
   * @param angle the angle in degree.
   */
  public void turnRight(double angle) {
    turnLeft(-angle);
  }
  
  /**
   * Turn the camera upward.
   * 
   * @param angle the angle in degree.
   */
  public void turnUpward(double angle) {
    if (angle != 0.0) {
      Iterator it = triangles.iterator();
      while (it.hasNext()) {
        ((Triangle)it.next()).turn(angle, UPDOWN);
      }
      
      turnCameraBase(-angle, UPDOWN);
    }
  }
  
  /**
   * Turn the camera downward.
   * 
   * @param angle the angle in degree.
   */
  public void turnDownward(double angle) {
    turnUpward(-angle);
  }
  
  /**
   * Move the camera forward.
   * 
   * @param step the distance how far the camera shall be moved.
   */
  public void moveForward(double step) {
    moveBackward(-step);
  }
  
  /**
   * Move the camera backward.
   * 
   * @param step the distance how far the camera shall be moved.
   */
  public void moveBackward(double step) {
    if (step != 0.0) {
      if (cameraPos[Z] - step * cameraY[Z] < 0.0)
        step = cameraPos[Z] / cameraY[Z];
      
      Iterator it = triangles.iterator();
      while (it.hasNext())
        ((Triangle)it.next()).move(step, Y);
      
      cameraPos[X] -= step * cameraY[X];
      cameraPos[Y] -= step * cameraY[Y];
      cameraPos[Z] -= step * cameraY[Z];
    }
  }
  
  /**
   * Move the camera to the left.
   * 
   * @param step the distance how far the camera shall be moved.
   */
  public void moveLeft(double step) {
    if (step != 0.0) {
      if (cameraPos[Z] - step * cameraX[Z] < 0.0)
        step = cameraPos[Z] / cameraX[Z];
      
      Iterator it = triangles.iterator();
      while (it.hasNext())
        ((Triangle)it.next()).move(step, X);
      
      cameraPos[X] -= step * cameraX[X];
      cameraPos[Y] -= step * cameraX[Y];
      cameraPos[Z] -= step * cameraX[Z];
    }
  }
  
  /**
   * Move the camera to the right.
   * 
   * @param step the distance how far the camera shall be moved.
   */
  public void moveRight(double step) {
    moveLeft(-step);
  }
  
  /**
   * Move the camera upward.
   * 
   * @param step the distance how far the camera shall be moved.
   */
  public void moveUpward(double step) {
    moveDownward(-step);
  }
  
  /**
   * Move the camera downward.
   * 
   * @param step the distance how far the camera shall be moved.
   */
  public void moveDownward(double step) {
    if (step != 0.0) {
      if (cameraPos[Z] - step * cameraZ[Z] < 0.0)
        step = cameraPos[Z] / cameraZ[Z];
      
      Iterator it = triangles.iterator();
      while (it.hasNext())
        ((Triangle)it.next()).move(step, Z);
      
      cameraPos[X] -= step * cameraZ[X];
      cameraPos[Y] -= step * cameraZ[Y];
      cameraPos[Z] -= step * cameraZ[Z];
    }
  }
  
  /**
   * Reset the camera direction to the default direction.
   */
  private void resetCamera() {
    double[] n = new double[DIMENSION];
    double dotProduct, alpha = 0.0;
    
    if (cameraX[X] == DEFAULT_CAMERA_X[X] &&
        cameraX[Y] == DEFAULT_CAMERA_X[Y] &&
        cameraX[Z] == DEFAULT_CAMERA_X[Z] &&
        cameraY[X] == DEFAULT_CAMERA_Y[X] &&
        cameraY[Y] == DEFAULT_CAMERA_Y[Y] &&
        cameraY[Z] == DEFAULT_CAMERA_Y[Z])
      return;
    
    if (cameraX[Z] != 0.0) {
      // z- and y-axes of camera coordinate and z-axis of world coordinate
      // are not coplanar, z-axes of camera coordinate and world coordinate
      // are not parallel.
      // -> turn scene left/right, so all three axes are coplanar
      
      /* Scene has to be turned by the angle between y-axis of the camera
       * coordinate and the plane defined by the z-axes of camera coordinate
       * and world coordinate.
       * 
       * z:     z-axis of the world coordinate (0;0;1)
       * z':    z-axis of the camera coordinate
       * y':    y-axis of the camera coordinate
       * E:     plane defined by z and z'
       * n:     normal vector of E
       * alpha: angle between y and E
       * 
       *    n = z x z'
       *        (z[Y] * z'[Z] - z[Z] * z'[Y])
       * => n = (z[Z] * z'[X] - z[X] * z'[Z])
       *        (z[X] * z'[Y] - z[Y] * z'[X])
       *        (0 * z'[Z] - 1 * z'[Y])
       * => n = (1 * z'[X] - 0 * z'[Z])
       *        (0 * z'[Y] - 0 * z'[X])
       *        (-z'[Y])
       * => n = ( z'[X])
       *        (   0  )
       * 
       *                        |y * n|
       *    cos(90 - alpha) = ---------
       *                       |y| * |n|
       *                          |y * n|
       * => alpha = 90 - arccos ---------
       *                         |y| * |n|
       */
      n[X] = -cameraZ[Y];
      n[Y] = cameraZ[X];
      n[Z] = 0.0;
        
      dotProduct = cameraY[X] * n[X] + cameraY[Y] * n[Y] + cameraY[Z] * n[Z];
      if (dotProduct < 0.0)
        dotProduct = -dotProduct;
        
      alpha = 90.0 - acos(dotProduct / Math.sqrt(cameraY[X] * cameraY[X] +
                                                 cameraY[Y] * cameraY[Y] +
                                                 cameraY[Z] * cameraY[Z]) /
                          Math.sqrt(n[X] * n[X] + n[Y] * n[Y]));
      // If x- and y-axes of the camera coordinate have the same signedness,
      // the scene has to be turned right, otherwise left.
      if (sgn(cameraX[Z]) == sgn(cameraY[Z]))
        alpha = -alpha;
    }
    
    turnLeft(alpha);
    
    // At this point, the x-axis of the camera coordinate is coplanar to the
    // xy-plane of the world coordinate. The y-axis of the camera coordinate
    // may not necessarily be coplanar to the xy-plane of the world coordinate.
    
    alpha = 0.0; // reset alpha
    
    if (cameraY[Z] != 0.0) {
      // y-axis of camera coordinate and xy-plane of world coordinate aren't
      // coplanar.
      // -> turn scene up-/downward, so axis and plane are coplanar
      
      /* y':    y-axis of the camera coordinate
       * alpha: angle between y' and its projection into the xy-plane
       * 
       *                 y'[Z]
       *    sin(alpha) = -----
       *                  |y'|
       *                       y'[Z]
       * => cos(90 - alpha) = -----
       *                        |y'|
       *                         y'[Z]
       * => alpha = 90 - arccos -----
       *                          |y'|
       * 
       * alpha is always the smallest angle between the y-axis of the camera
       * coordinate and the xy-plane (always out of [-90;90]).
       * If the z-coordinate of the camera's x-axis is negative, then the scene
       * must be turned upward, else downward.
       */
      alpha = 90.0 - acos(cameraY[Z] / Math.sqrt(cameraY[X] * cameraY[X] +
                                                 cameraY[Y] * cameraY[Y] +
                                                 cameraY[Z] * cameraY[Z]));
    }
    
    turnDownward(alpha);
    
    // At this point, the x- and y-axes of the camera coordinate are coplanar to
    // the xy-plane of the world coordinate. However, the z-axes of camera and
    // world coordinate may be antiparallel (world upside down ;). If this is
    // true, the scene has to be turned 180 upwards.
    if (cameraZ[X] == 0.0 && cameraZ[Y] == 0.0 && cameraZ[Z] < 0.0)
      turnUpward(180.0);
    
    // At this point, the x- and y-axes of the camera coordinate are coplanar to
    // the xy-plane of the world coordinate and the z-axes are parallel to each
    // other. However, the y-axes of camera and world coordinate may not yet be
    // parallel.
    
    alpha = 0.0; // reset alpha
    
    if (cameraY[X] != 0.0 || cameraY[Y] <= 0.0) {
      // y-axes of camera and world coordinate are not parallel.
      // -> turn scene horizontally
      alpha = Math.atan2(cameraY[Y], cameraY[X]) / Math.PI * 180.0;
      if (cameraY[X] > 0.0 || cameraY[Y] >= 0.0)
        alpha -= 90.0;
      else if (cameraY[Y] < 0.0)
        alpha += 270.0;
    }
    
    turnLeft(-alpha);
  }
  
  /**
   * Returns the arc cosine of a given value.
   * <p>
   * <table><tr>
   * <td>Pre:  <td>|d| &lt;= 1
   * </tr><tr>
   * <td>Post: <td>the returned angle is out of [0;180]
   * </tr></table> 
   * 
   * @param value a value.
   * @return the arc cosine of the argument or NaN if the argument is not out
   *         of [-1;1].
   */
  private double acos(double value) {
    return Math.acos(value) / Math.PI * 180.0;
  }
  
  /**
   * Return the algebraic signdness of the given value.
   * <p>
   * <table><tr>
   * <td>Pre:  <td>true
   * </tr><tr>
   * <td>Post: <td>true
   * </tr></table>
   * 
   * @param d the value, whose signedness should be evaluated.
   * @return the signedness of the given value.
   */
  private int sgn(double d) {
    return (d >= 0) ? 1 : -1;
  }
  
  /**
   * Turn the camera coordinate at given angle <i>angle</i> in direction
   * <i>direction</i>. If <i>direction</i> is UPDOWN, the coordinate is turned
   * around the z-axis of the camera coordinate. If direction is LEFTRIGHT,
   * the coordinate is turned around the x-axis of the camera coordinate.
   * <p>
   * <table><tr>
   * <td>Pre:  <td>direction is either <i>UPDOWN</i> or <i>LEFTRIGHT</i>.
   * </tr><tr>
   * <td>Post: <td>camera coordinate is turned at <i>angle</i> degree in
   *               direction.
   * </tr></table>
   * 
   * @param angle     The angle in degree.
   * @param direction The direction (horizontal or vertical) in which the
   *                  coordinate shall be turned.
   * @throws IllegalArgumentException if <i>direction</i> is neither
   *                                  <i>UPDOWN</i> nor LEFTRIGHT</i>.
   */
  private void turnCameraBase(double angle, int direction)
    throws IllegalArgumentException {
    Triangle base;
    double[] point, currentCameraX, currentCameraY, currentCameraZ;
    int n;
    
    if (direction != UPDOWN && direction != LEFTRIGHT)
      throw new IllegalArgumentException();
    
    base = new Triangle(1.0, 0.0, 0.0,
                        0.0, 1.0, 0.0,
                        0.0, 0.0, 1.0);
    currentCameraX = new double[DIMENSION];
    currentCameraY = new double[DIMENSION];
    currentCameraZ = new double[DIMENSION];
    
    // backup current base of the camera coordinate
    for (n = 0; n < DIMENSION; ++n) {
      currentCameraX[n] = cameraX[n];
      currentCameraY[n] = cameraY[n];
      currentCameraZ[n] = cameraZ[n];
    }
    
    base.turn(angle, direction);
    
    /* Calculate new base of the camera coordinate.
     * Transformation matrix T is matrix C formed by the base vectors of the
     * current camera coordinate.
     * 
     * E = world coordinate (identity matrix)
     * C = camera coordinate
     * T = transformation matrix
     * c = base vector in C
     * e = base vector in E
     * 
     * T = E' * C = E * C = C
     * 
     *         ( cameraX[X] cameraY[X] cameraZ[X] )
     * T = C = ( cameraX[Y] cameraY[Y] cameraZ[Y] )
     *         ( cameraX[Z] cameraY[Z] cameraZ[Z] )
     *
     * e = T * c = C * c
     */ 
    point = base.getPointA();
    cameraX[X] = currentCameraX[X] * point[X] +
                 currentCameraY[X] * point[Y] +
                 currentCameraZ[X] * point[Z];
    cameraX[Y] = currentCameraX[Y] * point[X] +
                 currentCameraY[Y] * point[Y] +
                 currentCameraZ[Y] * point[Z];
    cameraX[Z] = currentCameraX[Z] * point[X] +
                 currentCameraY[Z] * point[Y] +
                 currentCameraZ[Z] * point[Z];
  
    point = base.getPointB();
    cameraY[X] = currentCameraX[X] * point[X] +
                 currentCameraY[X] * point[Y] +
                 currentCameraZ[X] * point[Z];
    cameraY[Y] = currentCameraX[Y] * point[X] +
                 currentCameraY[Y] * point[Y] +
                 currentCameraZ[Y] * point[Z];
    cameraY[Z] = currentCameraX[Z] * point[X] +
                 currentCameraY[Z] * point[Y] +
                 currentCameraZ[Z] * point[Z];
  
    point = base.getPointC();
    cameraZ[X] = currentCameraX[X] * point[X] +
                 currentCameraY[X] * point[Y] +
                 currentCameraZ[X] * point[Z];
    cameraZ[Y] = currentCameraX[Y] * point[X] +
                 currentCameraY[Y] * point[Y] +
                 currentCameraZ[Y] * point[Z];
    cameraZ[Z] = currentCameraX[Z] * point[X] +
                 currentCameraY[Z] * point[Y] +
                 currentCameraZ[Z] * point[Z];
  }

}
