/*
 * Triangle.java
 */
 
package de.ba_ravensburg.gdv.model;

import java.awt.Color;

/**
 * Instances of this class represent three dimensional triangles. The triangles
 * are determined by three points in the coordinate space and their color.
 * 
 * @author Ansgar Wiechers
 */
public class Triangle {
  
  /**
   * Constant representing horizontal turns.
   */
  public static final int HORIZONTAL = Scene.LEFTRIGHT;
  
  /**
   * Constant representing vertical turns.
   */
  public static final int VERTICAL = Scene.UPDOWN;
  
  /**
   * The default color of a triangle.
   */
  static final Color DEFAULT_COLOR = new Color(255, 255, 255);
  
  private final int DIMENSION = Scene.DIMENSION;
  private final int X = Scene.X;
  private final int Y = Scene.Y;
  private final int Z = Scene.Z;
  private final int LEFTRIGHT = Scene.LEFTRIGHT;
  private final int UPDOWN = Scene.UPDOWN;
  
  private double[] pointA, pointB, pointC;
  private Color color;
  
  /**
   * Construction without arguments is not allowed, therefore the default
   * constructor is hidden.
   */
  private Triangle() {}
  
  /**
   * Constructs a new triangle by three given points. The color is set to the
   * default color.
   * 
   * @param xA x-coordinate of point A.
   * @param yA y-coordinate of point A.
   * @param zA z-coordinate of point A.
   * @param xB x-coordinate of point B.
   * @param yB y-coordinate of point B.
   * @param zB z-coordinate of point B.
   * @param xC x-coordinate of point C.
   * @param yC y-coordinate of point C.
   * @param zC z-coordinate of point C.
   * @throws IllegalArgumentException if at least two of the three points are
   *                                  equal.
   */
  public Triangle(double xA, double yA, double zA,
                  double xB, double yB, double zB,
                  double xC, double yC, double zC)
    throws IllegalArgumentException {
    this(xA, yA, zA, xB, yB, zB, xC, yC, zC, DEFAULT_COLOR);
  }
  
  /**
   * Constructs a new triangle by three given points and given color.
   * 
   * @param xA x-coordinate of point A.
   * @param yA y-coordinate of point A.
   * @param zA z-coordinate of point A.
   * @param xB x-coordinate of point B.
   * @param yB y-coordinate of point B.
   * @param zB z-coordinate of point B.
   * @param xC x-coordinate of point C.
   * @param yC y-coordinate of point C.
   * @param zC z-coordinate of point C.
   * @param theColor a color.
   * @throws IllegalArgumentException if at least two of the three points are
   *                                  equal.
   */
  public Triangle(double xA, double yA, double zA,
                  double xB, double yB, double zB,
                  double xC, double yC, double zC,
                  Color theColor)
    throws IllegalArgumentException {
    double[] a = {xA, yA, zA};
    double[] b = {xB, yB, zB};
    double[] c = {xC, yC, zC};
    
    if (isEqual(a, b) || isEqual(a, c) || isEqual(b, c))
      throw new IllegalArgumentException();
    
    pointA = a;
    pointB = b;
    pointC = c;
    color = theColor;
  }

  /**
   * Get the current color of the triangle.
   * <p>
   * <table><tr>
   * <td>Pre:  <td>true
   * </tr><tr>
   * <td>Post: <td>true
   * </tr></table>
   * 
   * @return current color of the triangle.
   */
  public Color getColor() {
    return color;
  }

  /**
   * Get point A of the triangle.
   * <p>
   * <table><tr>
   * <td>Pre:  <td>true
   * </tr><tr>
   * <td>Post: <td>true
   * </tr></table>
   * 
   * @return point A.
   */
  public double[] getPointA() {
    return pointA;
  }

  /**
   * Get point B of the triangle.
   * <p>
   * <table><tr>
   * <td>Pre:  <td>true
   * </tr><tr>
   * <td>Post: <td>true
   * </tr></table>
   * 
   * @return point B.
   */
  public double[] getPointB() {
    return pointB;
  }

  /**
   * Get point C of the triangle.
   * <p>
   * <table><tr>
   * <td>Pre:  <td>true
   * </tr><tr>
   * <td>Post: <td>true
   * </tr></table>
   * 
   * @return point C.
   */
  public double[] getPointC() {
    return pointC;
  }

  /**
   * Get the weighted distance of the triangle from the camera.
   * 
   * @return weighted distance of the triangle from the camera.
   */
  public double getDistance() {
    double[] max;
    
    max = (pointA[Y] > pointB[Y]) ? pointA : pointB;
    if (pointC[Y] > max[Y])
      max = pointC;
    
    return max[Y];
  }
  
  /**
   * Get the spherical coordinate angle of point A. Whether the <i>VERTICAL</i>
   * or <i>HORIZONTAL</i> angle should be obtained is specified by the parameter
   * <i>direction</i>.
   * <p>
   * <table><tr>
   * <td>Pre:  <td><i>direction</i> is either <i>HORIZONTAL</i> or
   *               <i>VERTICAL</i>.
   * </tr><tr>
   * <td>Post: <td>the spherical coordinate angle (<i>HORIZONTAL</i> or
   *               <i>VERTICAL</i>) is returned.
   * </tr></table>
   * 
   * @param direction the direction of the angle.
   * @return the spherical angle of point A in the given <i>direction</i>.
   * @throws IllegalArgumentException if <i>direction</i> is neither
   *                                  <i>HORIZONTAL</i> nor <i>VERTICAL</i>.
   */
  public double getAngleOfA(int direction) throws IllegalArgumentException {
    return getAngleOf(pointA, direction);
  }

  /**
   * Get the spherical coordinate angle of point B. Whether the <i>VERTICAL</i>
   * or <i>HORIZONTAL</i> angle should be obtained is specified by the parameter
   * <i>direction</i>.
   * <p>
   * <table><tr>
   * <td>Pre:  <td><i>direction</i> is either <i>HORIZONTAL</i> or
   *               <i>VERTICAL</i>.
   * </tr><tr>
   * <td>Post: <td>the spherical coordinate angle (<i>HORIZONTAL</i> or
   *               <i>VERTICAL</i>) is returned.
   * </tr></table>
   * 
   * @param direction the direction of the angle.
   * @return the spherical angle of point B in the given <i>direction</i>.
   * @throws IllegalArgumentException if <i>direction</i> is neither
   *                                  <i>HORIZONTAL</i> nor <i>VERTICAL</i>.
   */
  public double getAngleOfB(int direction) throws IllegalArgumentException {
    return getAngleOf(pointB, direction);
  }

  /**
   * Get the spherical coordinate angle of point C. Whether the <i>VERTICAL</i>
   * or <i>HORIZONTAL</i> angle should be obtained is specified by the parameter
   * <i>direction</i>.
   * <p>
   * <table><tr>
   * <td>Pre:  <td><i>direction</i> is either <i>HORIZONTAL</i> or
   *               <i>VERTICAL</i>.
   * </tr><tr>
   * <td>Post: <td>the spherical coordinate angle (<i>HORIZONTAL</i> or
   *               <i>VERTICAL</i>) is returned.
   * </tr></table>
   * 
   * @param direction the direction of the angle.
   * @return the spherical angle of point C in the given <i>direction</i>.
   * @throws IllegalArgumentException if <i>direction</i> is neither
   *                                  <i>HORIZONTAL</i> nor <i>VERTICAL</i>.
   */
  public double getAngleOfC(int direction) throws IllegalArgumentException {
    return getAngleOf(pointC, direction);
  }

  /**
   * Calculate the polar coordinate angle of a given point in <i>HORIZONTAL</i>
   * or <i>VERTICAL</i> direction.
   * <p>
   * <table><tr>
   * <td>Pre:  <td><i>direction</i> is either <i>HORIZONTAL</i> or
   *               <i>VERTICAL</i>.
   * </tr><tr>
   * <td>Post: <td>the spherical coordinate angle (<i>HORIZONTAL</i> or
   *               <i>VERTICAL</i>) is returned.
   * </tr></table>
   * 
   * @param point     a point.
   * @param direction the direction of the angle.
   * @return the spherical angle of the given <i>point</i> in the given
   *         <i>direction</i>.
   * @throws IllegalArgumentException if <i>direction</i> is neither
   *                                  <i>HORIZONTAL</i> nor <i>VERTICAL</i>.
   */
  private double getAngleOf(double[] point, int direction)
    throws IllegalArgumentException {
    double phi = 0.0;
    
    if (direction != HORIZONTAL && direction != VERTICAL)
      throw new IllegalArgumentException();
    
    if (direction == HORIZONTAL) {
      /* point[X] = 0,  point[Y] = 0  => phi = 0 (default value)
       * point[X] = 0,  point[Y] > 0  => phi = 0 (default value)
       * point[X] = 0,  point[Y] < 0  => phi = 180
       * point[X] < 0,  point[Y] = 0  => phi = 90
       * point[X] > 0,  point[Y] = 0  => phi = -90
       * point[X] != 0, point[Y] != 0 => phi = atan(point[Y]/point[X])
       */
      if (point[X] == 0.0) {
        if (point[Y] < 0.0)
          phi = 180.0;
      } else {
        if (point[Y] != 0) {
          phi = Math.atan2(point[Y], point[X]) / Math.PI * 180.0;
          // Correct phi, because it must be relative to the y-axis, not the
          // x-axis.
          if (point[X] > 0.0 || point[Y] > 0.0)
            phi -= 90.0;
          else
            phi += 270.0;
        } else {
          phi = 90.0 * ((point[X] < 0) ? 1 : -1);
        }
      }
    } else {
      /* point[Y] = 0,  point[Z] = 0  => phi = 0 (default value)
       * point[Y] > 0,  point[Z] = 0  => phi = 0 (default value)
       * point[Y] < 0,  point[Z] = 0  => phi = 180
       * point[Y] = 0,  point[Z] > 0  => phi = 90
       * point[Y] = 0,  point[Z] < 0  => phi = -90
       * point[Y] != 0, point[Z] != 0 => phi = atan(point[Y]/point[X])
       */
      if (point[Z] == 0.0) {
        if (point[Y] < 0.0)
          phi = 180.0;
      } else {
        if (point[Y] != 0) {
          phi = Math.atan2(point[Z], point[Y]) / Math.PI * 180.0;
        } else {
          phi = 90.0 * ((point[Z] < 0) ? -1 : 1);
        }
      }
    }
    
    return phi;
  }

  /**
   * Set new color of the triangle.
   * <p>
   * <table><tr>
   * <td>Pre:  <td>true
   * </tr><tr>
   * <td>Post: <td>true
   * </tr></table>
   * 
   * @param theColor new color to set.
   */
  public void setColor(Color theColor)  {
    color = theColor;
  }
  
  /**
   * Set point A of the triangle.
   * <p>
   * <table><tr>
   * <td>Pre:  <td>true
   * </tr><tr>
   * <td>Post: <td>pointA is (x;y;z)
   * </tr></table>
   * 
   * @param x x-coordinate of point A.
   * @param y y-coordinate of point A.
   * @param z z-coordinate of point A.
   */
  public void setPointA(double x, double y, double z) {
    pointA = new double[] {x, y, z};
  }

  /**
   * Set point B of the triangle.
   * <p>
   * <table><tr>
   * <td>Pre:  <td>true
   * </tr><tr>
   * <td>Post: <td>pointB is (x;y;z)
   * </tr></table>
   * 
   * @param x x-coordinate of point B.
   * @param y y-coordinate of point B.
   * @param z z-coordinate of point B.
   */
  public void setPointB(double x, double y, double z) {
    pointB = new double[] {x, y, z};
  }

  /**
   * Set point C of the triangle.
   * <p>
   * <table><tr>
   * <td>Pre:  <td>true
   * </tr><tr>
   * <td>Post: <td>pointC is (x;y;z)
   * </tr></table>
   * 
   * @param x x-coordinate of point C.
   * @param y y-coordinate of point C.
   * @param z z-coordinate of point C.
   */
  public void setPointC(double x, double y, double z) {
    pointC = new double[] {x, y, z};
  }
  
  /**
   * Turn the triangle by given angle in given plane around the camera position.
   * <p>
   * <table><tr>
   * <td>Pre:  <td>true
   * </tr><tr>
   * <td>Post: <td>the triangle is turned by <i>angle</i> degree.
   * </tr></table>
   *  
   * @param angle an angle.
   * @param plane plane in which the triangle should be turned.
   * @throws IllegalArgumentException if <i>plane</i> is neither <i>UPDOWN</i>
   *                                  nor <i>LEFTRIGHT</i>.
   */
  void turn(double angle, int plane) throws IllegalArgumentException {
    double x, y, z;
    
    if (plane != LEFTRIGHT && plane != UPDOWN)
      throw new IllegalArgumentException();
    
    switch (plane) {
      case LEFTRIGHT:
        // turn pointA
        x = pointA[X];
        y = pointA[Y];
        pointA[X] = cos(angle) * x + sin(angle) * y;
        pointA[Y] = cos(angle) * y - sin(angle) * x;
        
        // turn pointB
        x = pointB[X];
        y = pointB[Y];
        pointB[X] = cos(angle) * x + sin(angle) * y;
        pointB[Y] = cos(angle) * y - sin(angle) * x;
        
        // turn pointC
        x = pointC[X];
        y = pointC[Y];
        pointC[X] = cos(angle) * x + sin(angle) * y;
        pointC[Y] = cos(angle) * y - sin(angle) * x;
        
        break;
        
      case UPDOWN:
        // turn pointA
        y = pointA[Y];
        z = pointA[Z];
        pointA[Y] = cos(angle) * y + sin(angle) * z;
        pointA[Z] = cos(angle) * z - sin(angle) * y;
        
        // turn pointB
        y = pointB[Y];
        z = pointB[Z];
        pointB[Y] = cos(angle) * y + sin(angle) * z;
        pointB[Z] = cos(angle) * z - sin(angle) * y;
        
        // turn pointC
        y = pointC[Y];
        z = pointC[Z];
        pointC[Y] = cos(angle) * y + sin(angle) * z;
        pointC[Z] = cos(angle) * z - sin(angle) * y;
        
        break;
    }
  }
  
  /**
   * Move the triangle by given amount step into direction.
   * <p>
   * <table><tr>
   * <td>Pre:  <td>true
   * </tr><tr>
   * <td>Post: <td>the triangle is moved <i>step</i> units into direction
   *               <i>direction</i>.
   * </tr></table>
   * 
   * @param step      how far to move the triangle.
   * @param direction axis along which the triangle should be moved.
   * @throws IllegalArgumentException if <i>direction</i> is neither of
   *                                  <i>X</i>, <i>Y</i> or <i>Z</i>.
   */
  void move(double step, int direction) throws IllegalArgumentException {
    if (direction != X && direction != Y && direction != Z)
      throw new IllegalArgumentException();
    
    pointA[direction] += step;
    pointB[direction] += step;
    pointC[direction] += step;
  }
  
  /**
   * Determine whether two points are equal (IOW have the same corresponding
   * coordinates.
   * <p>
   * <table><tr>
   * <td>Pre:  <td>point1.length == point2.length == DIMENSION
   * </tr><tr>
   * <td>Post: <td>point1.length == point2.length == DIMENSION
   * </tr></table>
   * 
   * @param point1 First point to compare.
   * @param point2 Second point to compare.
   * @return true if point1[n] == point2[n] for each n out of [1..DIMENSION],
   *         else false. 
   */
  private boolean isEqual(double[] point1, double[] point2) {
    int n;
    
    for (n = 0; n < DIMENSION; ++n)
      if (point1[n] != point2[n])
        return false;
    
    return true;
  }
  
  /**
   * Returns the trigonometric sine of an angle.
   * <p>
   * <table><tr>
   * <td>Pre:  <td>true
   * </tr><tr>
   * <td>Post: <td>true
   * </tr></table>
   * 
   * @param angle an angle, in degrees.
   * @return the sine of the argument.
   */
  private double sin(double angle) {
  return Math.sin(angle / 180.0 * Math.PI);
  }
  
  /**
   * Returns the trigonometric cosine of an angle.
   * <p>
   * <table><tr>
   * <td>Pre:  <td>true
   * </tr><tr>
   * <td>Post: <td>true
   * </tr></table>
   * 
   * @param angle an angle, in degrees.
   * @return the cosine of the argument.
   */
  private double cos(double angle) {
    return Math.cos(angle / 180.0 * Math.PI);
  }

}
